#include "PCCond.h"


int timeelapsed = 0;			//global value representing number of timer "ticks"

/* FUNCTION: OnMultimediaTimer
 * PURPOSE: This function is called from a separate thread every X ms    *
 * (as specified in the timeSetEvent function call). It increments a     *
 * global variable (timeelapsed) whose current value can be used to      *
 * calculate how much time has passed (timeelapsed [tick] * X [ms/tick] )*
 * since the timeSetEvent function was called.                           *
 * DATE: 5/22/02                                                         *
 * CONSTANTS: none                                                       *
 * MODULES: PCCond                                                       */
void CALLBACK OnMultimediaTimer( UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2 )
{
	timeelapsed++;
}


/* FUNCTION: writeLittleEndian                                                     *
 * PURPOSE: Given a 16bit value and a location in a character array, this function *
 * will write the value into the array in Little-Endian format.                    *
 * DATE: 5/22/02                                                                   *
 * CONSTANTS: none                                                                 *
 * MODULES: PCCond                                                                 */
void writeLittleEndian( char *writeto, short data )
{
	writeto[0] = *((char*)&data + 1);
	writeto[1] = *((char*)&data);
}


/* FUNCTION: copyPCToPalm
 * PURPOSE: Given the output of the runCPI/runPROB functions (ie. the missing hash     *
 * values from the PalmPilot and from the PC), a character array and a pointer to a    *
 * array of memos on the PC, this function will put the data strings corresponding     *
 * to hashes that the PalmPilot is missing in the character array in the following     *
 * format:                                                                             *
 * [#memos_in_array(16bit)][memo 1 string(variable length, null term)]                 *
 * [memo 1 hash (16bit)] [memo 2 string][memo 2 hash] ... [memo X string][memo X hash] *
 * Note - This data cannot exceed in size the constant REMOTEDATASIZE!                 *
 * The return value is simply a pointer to the character array, AT THE END OF ALL THE  *
 * DATA INSERTED!                                                                      *
 * DATE: 5/22/02                                                                       *
 * CONSTANTS: none                                                                     *
 * MODULES: PCCond                                                                     */
char *copyPCToPalm( int **missingdata, char *remotedata, memo *memodata)
{
	char *dataoffsetptr;
	
	dataoffsetptr = remotedata + 2;				//keep an open space to write the # of memos placed in the array

	if( memodata[0].hash == 0 )					//if there are no memos in the PC file
	{											//then there can be no missing memos for the Palm to take
		writeLittleEndian( remotedata, 0 );
		return( remotedata + 2 );
	}

	for( int x = 0; x < missingdata[0][0]; x++ )	//for each Palm missing hash value
	{												//get the memo, put it in the array
		strcpy( dataoffsetptr, getStringWithHash( missingdata[0][x+1], memodata ));
		dataoffsetptr += strlen(dataoffsetptr)+1;
		writeLittleEndian( dataoffsetptr, (short)missingdata[0][x+1] );	//and the memo's
		dataoffsetptr += 2;												//hash value too.
	}
	writeLittleEndian( remotedata, (short)missingdata[0][0] );	//when we finish write the # of memos placed in the array

	return(dataoffsetptr);						//return the pointer that is PAST the data inserted
}


/* FUNCTION: isNotIn                                                                *
 * PURPOSE: Given a number to look for and an array (with its size reported in      *
 * the 0'th index) to look in, this function will scan linearly through the array   *
 * and return true if that number is found and false if it is not.                  *
 * DATE: 5/22/02                                                                    *
 * CONSTANTS: none                                                                  *
 * MODULES: PCCond                                                                  *
 * SUGGESTIONS: Remove this function if it is used only once in the program.        */
bool isNotIn( int lookfor, int *lookin )
{
	for( int x = 0; x < lookin[0]; x++ )	//go through reported array size
		if( lookfor == lookin[x+1] )
			return(false);
	return(true);
}


/* FUNCTION: getHH                                                                     *
 * PURPOSE: Given the handle to the pre-openend PalmPilot sample polynomials database, *
 * and an integer containing the method of synchronization used, this function will    *
 * download the appropriate records containing the sample polynomial points themselves *
 * and will put them into an array (with the size of the array as the 0th index) which *
 * it will return. It will also set the integer pointer argument to the number of      *
 * records actually downloaded.                                                        *
 * DATE: 5/22/02                                                                       *
 * CONSTANTS: PROB1                                                                    *
 *            DIFFERNECES                                                              *
 * MODULES: PCCond                                                                     *
 * SUGGESTIONS: Why should we bother with the numrecs pointer, when that data is already*
 * in the array anyway?                                                                */
int *getHH( BYTE db, int *numrecs, int synctype )
{
	int *hhpoly, temp, diff;
	CRawRecordInfo rawrecord;

	if( synctype != 0 )					//depending on the type of sync, we will be downloading different
		diff = PROB1;					//numbers of sample evaluations. We will only allocate enough space
	else								//for the amount that we download.
		diff = DIFFERENCES;

	hhpoly = new int[diff+1];			//allocate storage space for the sample polynomial values

	memset(&rawrecord, 0, sizeof(CRawRecordInfo));      //read the sample polys from the handheld
	rawrecord.m_pBytes = (BYTE*)new short[diff+1];
	rawrecord.m_FileHandle = db;
	rawrecord.m_RecIndex = synctype;
	rawrecord.m_TotalBytes = (diff+1) * sizeof(short);
	SyncReadRecordByIndex(rawrecord);

	*numrecs = (rawrecord.m_pBytes[0] << 8) | rawrecord.m_pBytes[1];	//first part of record contains # polys stored

	for( unsigned int x = 1; x < diff * sizeof( short ); x += 2 )	//convert each sample poly from Little-Endian and store it
	{
		temp = (rawrecord.m_pBytes[x+1] << 8) | rawrecord.m_pBytes[x+2];
		hhpoly[(x+1)/2] = temp;
	}
	hhpoly[0] = diff;
	delete( rawrecord.m_pBytes );

	return( hhpoly );
}



/* FUNCTION: rpcOnPalm                                                                                   *
 * PURPOSE: This function encapsulates the Remote Procedure Call (RPC) on the PalmPilot                  *
 * that is necessary to convert the PC's missing hash values to actual memo text that                    *
 * is stored on the PalmPilot, and to provide the PalmPilot with the actual memo text                    *
 * for its missing hashes. The function takes as arguments: the output of the                            *
 * runCPI/runProb function (that contains the PC's missing hashes), a pointer to the array               *
 * of memos on the PC, and an empty CCallModuleParams object that the function will                      *
 * manipulate, but will not deallocate. The data returned by this function are the indexes               *
 * (in the Palm's document database) of the memos corresponding to the missing hash values               *
 * found in the indexes of the second column of the missingdata array (ie. missingdata[1][x])            *
 * and the indexes/hash value pairs of any memos marked for deletion on the PalmPilot. This              *
 * returned data is in the following format:                                                             *
 * [#memos marked deleted on Palm (16bit)] [index of deleted memo1 in Palm doc db. (16bit)]              *
 * [hash value of deleted memo1 (16bit)] ... [indexX][hashX]                                             *
 * [#missing memos returned from palm (16bit)][index in Palm document database of missing memo1 (16bit)] *
 * [index in Palm document database of missing memo2]...[indx in Palm doc db. of missing memoX]          *
 * DATE: 5/22/02                                                                                         *
 * CONSTANTS: REMOTEDATASIZE                                                                             *
 *            'strt' - Creator ID of application to RPC                                                  *
 *            'appl' - dwTypeID?                                                                         *
 *            REPLYDATASIZE                                                                              *
 *            1 - was, at one point, used as a sentinel value to indicate end of remotedata array?       *
 * MODULE: PCCond                                                                                        *
 * SUGGESTION: Try and get rid of the memset function, since it is likely no longer needed.              */
short *rpcOnPalm( int **missingdata, memo *memodata, CCallModuleParams *params)
{
	char remotedata[REMOTEDATASIZE], *pastdata;
	long dummy;

	memset(remotedata, 1, sizeof(char)*REMOTEDATASIZE);	//probably was to set all data to ones so that after
	pastdata = remotedata;                              //the remote data is written, it will be ended with a 1

	writeLittleEndian( pastdata, missingdata[1][0] );	//get hash values to be converted to memo text by the Palm
	pastdata += 2;										//into the remote data array
	for( int y = 0; y < missingdata[1][0]; y++ )
	{
		writeLittleEndian( pastdata, (short)missingdata[1][y+1] );
		pastdata += 2;
	}

	pastdata = copyPCToPalm( missingdata, pastdata, memodata );	//get missing memo text and the hash values of the memos
                                                                //that the Palm is missing
	memset(params, 0, sizeof(CCallModuleParams));			//setup and call RPC on Palm
	params->m_dwCreatorID = 'strt';
	params->m_dwTypeID = 'appl';
	params->m_dwParamSize = sizeof(char) * (pastdata - remotedata);
	params->m_pParam = remotedata;
	params->m_pResultBuf = new short[REPLYDATASIZE];
	params->m_dwResultBufSize = sizeof(short) * REPLYDATASIZE;
	dummy = SyncCallRemoteModule(params);

	return((short*)params->m_pResultBuf);
}


/* FUNCTION: getPalmDeleted                                                                 *
 * PURPOSE: The PalmPilot has returned the index/hashvalue pairs of the memos it has        *
 * marked for deletion. Given that data, this function splits it into two distint arrays:   *
 * one of the deleted memos' hash values, and one of the deleted memos' indexes. These      *
 * arrays are "returned" through the second and third arguments.                            *
 * DATE: 5/22/02                                                                            *
 * CONSTANTS: none                                                                          *
 * MODULES: PCCond                                                                          */
void getPalmDeleted( short *palmreturned, int **palmdeleted, int **palmdeletedhash )
{
	int *palmdel = 0, *palmhash = 0;
	
	palmdel = new int[palmreturned[0]+1];		//create the two arrays (of the same size as the number of hash/index pairs)
	palmhash = new int[palmreturned[0]+1];
	palmdel[0] = palmreturned[0];				//explicitely store their size in the 0th index
	palmhash[0] = palmreturned[0];
	for( int x = 0; x < palmreturned[0]; x++ )	//then move through the deleted index/hash
	{                                           //pairs, and split them into the two arrays.
		palmdel[x+1] = palmreturned[2*x+1];
		palmhash[x+1] = palmreturned[2*x+2];
	}

	*palmdeleted = palmdel;						//simply reassign the passed pointers to the newly created arrays
	*palmdeletedhash = palmhash;
}


/* FUNCTION: synchronize
 * PURPOSE: This is the workhorse function to reconcile the PC and Palm called by the conduit dll.           *
 * It is the code for the conduit itself, although this it mostly just calls conduit helper functions        *
 * in the right order and transfers data between them. This is the order in which it works:                  *
 * 1) It downloads the sample polynomials from the remote machine                                            *
 * 2) It loads the local machine's memo file into memory and gets its hash values                            *
 * 3) It runs the reconciliation algorithm to identify missing hashes from the local and remote databases    *
 * 4) It prepares the memos that the remote machine is missing as well as the hashes that the local machine  *
 *    is missing and sends them to the remote machine.                                                       *
 * 5) It gets the memos from the remote machine corresponding to the hash values that were missing from the  *
 *    local machine. It encorporates those memos into the local machines database.                           *
 * DATE: 6/21/02                                                                                             *
 * CONSTANTS: NUMMEMOS                                                                                       *
 *            40 - maximum file name length                                                                  *
 *            120 - maximum path length                                                                      *
 *            200 - maximum filename + path length                                                           *
 *            20 - delay between each "tick" of the multimedia timer                                         *
 *            5 - granularity/resolution of the system clock                                                 *
 * MODULES: PCCond                                                                                           *
 * SUGGESTIONS: Why do we need "count" variable? (see getHH function)                                        */
void synchronize( char *localname, char *pathname, char **remotename, int *timervalues )
{
	int count;		
	
	//these pointers will be dynamically allocated to hold there respective values
	int *hhpoly, **missingdata, *palmdeleted = 0, *palmdeletedhash = 0, *pchash;
	short *palmreturn;
	
	int nummemos;										//keep track of how many memos are on the PC
	CCallModuleParams params;							//holds the RPC call parameters
	BYTE db;											//reference to database on the PalmPilot
	char fileName[40],path[120],pcMemoFilePath[200];	//strings for the PC files/paths
	memo memodata[NUMMEMOS];	//memory location to hold the memos from the PC's memo database

	//get the path to the PC's memo file into the pcMemoFilePath variable
	if (strlen(localname) != 0)			//if a local filename was provided use it
        strcpy(fileName,localname);
    else											//otherwise, "SyncMemo.dat" will be used.
		strcpy(fileName,"SyncMemo.dat");
	strcpy(path,pathname);					//path to user's directory
	strcpy(pcMemoFilePath, strcat(path,fileName));	
		
	timervalues[1] = timeelapsed;
	
	//Step (1): get the PalmPilot sample characteristic polynomial evaluations
	SyncOpenDB(remotename[0], (short)0, db, (BYTE)(eDbRead));

#ifdef CPISYNC	
	hhpoly = getHH( db, &count, 0 );		//0 for CPIsync
#endif
#ifdef PROBSYNC
	hhpoly = getHH( db, &count, 1 );		//1 for PROBsync
#endif

	timervalues[2] = timeelapsed;

	//Step (2): load the PC's memos into memory
	initializeNewData( memodata );					
	nummemos = loadFile( memodata, pcMemoFilePath);

	timervalues[3] = timeelapsed;
		
	//Step (2): put the PC's memo hashes in the dynamically allocated "pchash" array
	pchash = new int[nummemos+1];
	pchash[0] = nummemos;					//store the number of memos in the array as 0th element in it
	for( int x = 0; x < nummemos; x++ )
		pchash[x+1] = memodata[x].hash;

	timervalues[4] = timeelapsed;

	//Step (3): run the reconciliation algorithm (ONLY ONE OF THE FOLLOWING TWO LINES SHOULD BE UNCOMMENTED)
#ifdef CPISYNC
	missingdata = runCPI( pchash, hhpoly, count );			//for CPIsync
#endif
#ifdef PROBSYNC
	missingdata = runProb( pchash, &hhpoly, count, db );	//for PROBsync.
#endif

	timervalues[5] = timeelapsed;

	//clean up what you can...
	SyncCloseDB(db);			//close the samply polynomial DB on the PalmPilot
	delete(hhpoly);				//de-allocate that handheld polynomials evaluations

	if( missingdata == NULL )	//if we have received nothing from the algorithm, something has gone wrong
		return;

	timervalues[6] = timeelapsed;

	//Step (4): run the Remote Procedure Call (RPC) on the PalmPilot
	palmreturn = rpcOnPalm( missingdata, memodata, &params );

	timervalues[7] = timeelapsed;
	
	//Step (5): parse the returned data from the RPC - we want the deleted memos from the PalmPilot
	getPalmDeleted( palmreturn, &palmdeleted, &palmdeletedhash );

	timervalues[8] = timeelapsed;

	//Step (5): add any missing data to the PC
	writeMissingToPC2( missingdata, memodata, &nummemos, (char*)palmreturn, palmdeleted );
	
	timervalues[9] = timeelapsed;
	
	//Step (5): we should purge the deleted memos from the PalmPilot, but only if there are any
	if( palmdeleted[0] != 0 )
	{
		SyncOpenDB( "SyncDocDB", (short)0, db, (BYTE)(eDbRead | eDbWrite) );
		SyncPurgeDeletedRecs(db);
		SyncCloseDB(db);
	}

	//clean up what you can...
	delete(missingdata[0]);
	delete(missingdata[1]);
	delete(missingdata);
		
	//delete any PC memos that were deleted on the palm
	for( x = 0; x < palmdeletedhash[0]; x++ )
		if( !isNotIn( palmdeletedhash[x+1], pchash ) )
		{
			deleteRecord( memodata, palmdeletedhash[x+1], nummemos );
			nummemos--;
		}

	//clean up what you can...
	delete( pchash );
	delete( palmreturn );
	delete( palmdeleted );
	delete( palmdeletedhash );

	timervalues[10] = timeelapsed;

	//save the memos back to the PC's memo file, and clean up the memory they occupied
	saveFile( pcMemoFilePath, nummemos, memodata);
	freeData( memodata );

	timervalues[11] = timeelapsed;
}


/* FUNCTION: writeMissingToPC2                                                                                        *
 * PURPOSE: Given the PC's missing hash values, a pointer to the array of the PC's memos, the number of memos on      *
 * the PC, the text of the PC's missing memos returned from the Palm and the indexes of the memos on the Palm marked  *  
 * for deletion, this function will place the memo text returned from the Palm into the PC's array of memos.          *
 * DATE: 6/21/02                                                                                                      *
 * CONSTANTS: MEMOSIZE                                                                                                *
 * MODULES: PCCond                                                                                                    *
 * SUGGESTION: couldn't we remove the buffer variable?                                                                */
void writeMissingToPC2( int **missingdata, memo *memodata, int *nummemos, char *palmreturn, int *palmdeleted )
{
	char buffer[MEMOSIZE];						
	int numreturned;

	palmreturn += palmdeleted[0] * 4 + 2;		//skip past the deleted memos and hash values

	//add the PalmPilot returned memos into the PC's existing memo set
	numreturned = missingdata[1][0];			//get the number of memos missing from PC (presumably = # memos returned)
	for( int x = 0; x < numreturned; x++ )		
	{											//add to PC's memo array
		if( palmreturn[0] != '\0' )
		{
			strcpy( buffer, palmreturn );
			createMemo( buffer, missingdata[1][x+1], nummemos, memodata );			
		}
		palmreturn += strlen(palmreturn) + 1;	//skip to next memo in array
	}
}
