#include "\Common\Constants.h"

#ifdef HASH2				//only use this file if the AVLtree style hash database is desired in constants.h

#include "\Palm\PalmHashDb2\PalmHashDb2.h"


/* FUNCTION: blockInsertHashIndexes                                                                              *
 * PURPOSE: Given the number of hashes passed to the function, an array of those hash structures and a reference *
 * to the pre-opened hash database, this function will insert all the hash values in one call. The savings of    *
 * using this function over simply calling insertHashIndex multiple times is that this function only gets and    *
 * locks the record ONE time per batch of hashvalues. insertHashIndex gets and locks for each hash value inserted*
 * DATE: 7/20/02                                                                                                 *
 * CONSTANTS: NUMMEMOS                                                                                           *
 *            0 - the index of the record in the hash database that has all the hash values                      *
 * MODULES: PalmHashDb                                                                                           */
void blockInsertHashIndexes( UInt16 numhashes, HashRecordType hashvalues[], DmOpenRef hashDb )
{
	UInt16 x;
	MemHandle myrecord;
	AVLNode *avlarray;

	myrecord = DmGetRecord( hashDb, 0 );			//access the avl tree record (the hash record), ONLY DONE ONCE FOR ALL INSERTIONS
	avlarray = MemHandleLock( myrecord );
	
	for( x = 0; x < numhashes; x++ )
		doInsertHashIndex( hashvalues[x], hashDb, 1, avlarray );

	MemHandleUnlock( myrecord );					//unlock and release, ONLY DONE ONCE FOR ALL INSERTIONS	
	DmReleaseRecord( hashDb, 0, true ); 
}


/* FUNCTION: calcHashValue                                               *
 * PURPOSE: Given a pointer to a character string, this function will    *
 * return the 16bit hash value for it. The hash function implemented     *
 * is simply the xor(?) of each even byte for the least significant byte *
 * of the hash value, and the xor(?) of each odd byte for the most       *
 * significant byte of the hash value.                                   *
 * DATE: 8/16/02                                                         *
 * CONSTANTS: DIFFERENCES                                                *
 *            256 - MAX 8bit value                                       *
 *            32749 - Prime field                                        *
 * MODULES: PalmHashDb                                                   */
UInt16 calcHashValue( char *data )
{
	char hash[2] = {0,0};			//these two 8bit chars form the space for our 16bit hash value
	UInt16 hashval;
	int temp,x;

	temp = StrLen(data);
	if( temp % 2 )								//we must check the length so that we know whether the string 
	{
		for( x = 0; x < temp-1; x += 2 )		//ends on an odd or even character
		{
			hash[1] ^= data[x];
			hash[0] ^= data[x+1];
		}
		hash[1] ^= data[x];
	}
	else
	{
		for( x = 0; x < temp; x += 2 )
		{
			hash[1] ^= data[x];
			hash[0] ^= data[x+1];
		}
	}

	hashval = *(UInt16 *)hash;
	hashval = hashval % 32749;
	
	//avoid collision between hash value and sample points (collision is when sample-hash = 0 or sample-hash = prime field)
	//the sample points go as {0, -1, 1, -2, 2... -N/2} for N number of differences so adding half the value of N (or 
	//subtracting it) should avoid collisions
	if( hashval < DIFFERENCES/2 + 1 )
		hashval += DIFFERENCES/2 + 1;
	else if( hashval > 32749 - (DIFFERENCES/2 + 1) )
		hashval -= DIFFERENCES/2 + 1;
	
	return( hashval );
}


/* FUNCTION: doInsertHashIndex                                                                            *
 * PURPOSE: This function is the one that does the actual work of inserting hash indexes into the hash    *
 * database. Given a complete hashrecord structure, and a reference to the pre-opened hash database, this *
 * function will insert that hashrecord structure into the hash database. The method used here is to      *
 * store the hash entries in an AVL tree, placed on top of an array. This function also allows for a      *
 * "block" mode in which access to the hash records are provided by the calling function. see the         *
 * blockInsertHashIndexes function for more on "block" mode.                                              *
 * DATE: 7/20/02                                                                                          *
 * CONSTANTS: 0 - the index of the record in the hash database that has all the hash values               *
 * MODULES: PalmHashDb                                                                                    */
void doInsertHashIndex( HashRecordType hashdata, DmOpenRef hashDb, UInt16 isblockmode, AVLNode *preopenedarray )
{
	MemHandle myrecord;
	AVLNode *avlarray, temp, tempnodetemp, forward1temp, forward2temp;
	UInt16 head, firstfreeindex, newnode, currentnode, tempnode, forward1 = 0, forward2 = 0, temp1;

	UInt8 height = 0, casenum;

	if( !isblockmode )			//in "block" mode, the calling function takes care of getting the record and locking it
	{
		myrecord = DmGetRecord( hashDb, 0 );			//access the avl tree record (the hash record)
		avlarray = MemHandleLock( myrecord );
	}
	else
		avlarray = preopenedarray;
	
	head = avlarray[0].left;			//the index of the "head" node of the avl tree is stored in the zero'th index's "left" data member
	firstfreeindex = avlarray[0].right;	//the first free index in the array is stored in the zero'th index's "right" data member

	//if the tree is empty, we can just make the new node the head node
	if( head == 0 )
	{
		head = firstfreeindex++;			//start the head as the 1st index 
		temp.data = hashdata;
		temp.balance = 0;
		temp.left = 0;
		temp.right = 0;
		temp.prevnode = 0;
		DmWrite( avlarray, sizeof(AVLNode), &temp, sizeof( temp ) ); 
	}
	else		//if the tree is not empty, insert into the tree, making sure to keep the AVL properties
	{
		newnode = firstfreeindex++;
		temp.data = hashdata;
		temp.left = 0;
		temp.right = 0;
		temp.balance = 0;

		//First, insert the new node into the tree as if it were a standard binary tree
		currentnode = head;
		while( 1 )				
		{
			if( hashdata.hashval > avlarray[currentnode].data.hashval )		
				if( avlarray[currentnode].right != 0 )						//go to the right, if it exists
					currentnode = avlarray[currentnode].right;
				else												//if we reach the end in a direction we need to 
				{													//go, then thats where we add our new node
					DmWrite(avlarray, currentnode * sizeof(AVLNode) + 2*sizeof(UInt16), &newnode, sizeof(newnode) );
					break;
				}
			else
				if( avlarray[currentnode].left != 0 )						//go to the left, if it exists
					currentnode = avlarray[currentnode].left;
				else												//if we reach the end in a direction we need to
				{													//go, then thats where we add our new node
					DmWrite(avlarray, currentnode * sizeof(AVLNode) + 1*sizeof(UInt16), &newnode, sizeof(newnode) );
					break;
				}
		}
		temp.prevnode = currentnode;						//keep track of the reverse path
		DmWrite( avlarray, newnode * sizeof(AVLNode), &temp, sizeof( temp ) );

		//Second, update the balance indicators of the nodes in the reverse path of the new node.
		//Only go back until you have balanced a node (ie. 1 -> 0 or -1 -> 0) or completely unbalanced
		//it, (ie. -1 -> -2 or 1 -> 2). We will use the balance factors to assist in the rotation
		tempnode = newnode;			//start one behind the new node and work backwards
		do
		{
			if( tempnode == avlarray[avlarray[tempnode].prevnode].right )	//if we came from the right, then increment the balance
				temp1=avlarray[avlarray[tempnode].prevnode].balance+1;
			else															//if we came from the left, then decrement the balance
				temp1=avlarray[avlarray[tempnode].prevnode].balance-1;
			DmWrite(avlarray, avlarray[tempnode].prevnode * sizeof(AVLNode), &temp1, sizeof(temp1) );	//write the balance to the record
			
			forward2 = forward1;					//as we move up the tree, keep track of the last two
			forward1 = tempnode;					//nodes that we have visited, which are the nodes that
			tempnode = avlarray[tempnode].prevnode;	//are "in-front" of the temporary node as it moves back

			if( avlarray[tempnode].prevnode == 0 )	//if we have reached the root node, then stop
				break;

		} while( (avlarray[tempnode].balance != 0) && (avlarray[tempnode].balance != 2) && (avlarray[tempnode].balance != -2) );
	
		//Third, figure out if the tree is unbalanced, and if it is, then in what way is it unbalanced
		//There will be 4 cases, 1 pair (casenum = 1,4) that needs a "single" rotation, and one pair
		//(casenum = 2,3) that needs a "double" rotation. The cases that I use are described in "Data Structures
		//and Algorithms in C" by Mark Allen Weiss, 2nd ed, pg 111
		if(!( (avlarray[tempnode].balance != 2) && (avlarray[tempnode].balance != -2) ))		//if the node we stopped at is imbalanced
		{																						

			if( avlarray[tempnode].balance == -2 )
				if( avlarray[forward1].balance == -1 )
					casenum = 1;					//single rotation to the right
				else
					casenum = 2;					//double rotation left, right

			if( avlarray[tempnode].balance == 2 )
				if( avlarray[forward1].balance == 1 )
					casenum = 4;					//single rotation to the left
				else
					casenum = 3;					//double rotation right, left
	

			tempnodetemp = avlarray[tempnode];		//copy the current nodes into their temporary equivalents
			forward1temp = avlarray[forward1];		//so that we will make all the modifications to them and write 
			forward2temp = avlarray[forward2];		//them all at once to the database (rather than write each modification seperately)

			switch( casenum )
			{

			case 1:	//single rotation cases
			case 4: forward1temp.prevnode = avlarray[tempnode].prevnode;
					if( forward1temp.prevnode != 0 )
						if( avlarray[forward1temp.prevnode].right == tempnode )
							DmWrite( avlarray, forward1temp.prevnode * sizeof(AVLNode) + 2*sizeof(UInt16), &forward1, sizeof(UInt16) );
						else
							DmWrite( avlarray, forward1temp.prevnode * sizeof(AVLNode) + sizeof(UInt16), &forward1, sizeof(UInt16) );

					tempnodetemp.prevnode = forward1;

					if( casenum == 1)		//rotate to the right (case 1)
					{
						tempnodetemp.left = avlarray[forward1].right;
						if( tempnodetemp.left != 0 )
							DmWrite( avlarray, tempnodetemp.left * sizeof(AVLNode) + 3*sizeof(UInt16), &tempnode, sizeof(UInt16) );
						forward1temp.right = tempnode;
					}
					else					//rotate to the left (case 4)
					{
						tempnodetemp.right = avlarray[forward1].left;
						if( tempnodetemp.right != 0 )
							DmWrite( avlarray, tempnodetemp.right * sizeof(AVLNode) + 3*sizeof(UInt16), &tempnode, sizeof(UInt16) );
						forward1temp.left = tempnode;
					}

					temp1 = 0;
					tempnodetemp.balance = 0;
					forward1temp.balance = 0;

					DmWrite( avlarray, tempnode * sizeof(AVLNode), &tempnodetemp, sizeof(AVLNode) );
					DmWrite( avlarray, forward1 * sizeof(AVLNode), &forward1temp, sizeof(AVLNode) );
					if( tempnode == head )		//if we created a new head, (by replacing tempnode with forward1)
						head = forward1;		//then keep track of it
					break;

			case 2:	//double rotation cases
			case 3:	
					if( casenum == 2 )
					{
						forward1temp.right = avlarray[forward2].left;
						if( forward1temp.right != 0 )
							DmWrite( avlarray, forward1temp.right * sizeof(AVLNode) + 3*sizeof(UInt16), &forward1, sizeof(UInt16) );
						tempnodetemp.left = avlarray[forward2].right;
						if( tempnodetemp.left != 0 )
							DmWrite( avlarray, tempnodetemp.left * sizeof(AVLNode) + 3*sizeof(UInt16), &tempnode, sizeof(UInt16) );
						forward2temp.left = forward1;
						forward2temp.right = tempnode;
					}
					else
					{
						forward2temp.right = avlarray[tempnode].right;
						forward2temp.left = tempnode;
						tempnodetemp.right = avlarray[forward2].left;
						if( forward2temp.right != 0 )
							DmWrite( avlarray, forward2temp.right * sizeof(AVLNode) + 2*sizeof(UInt16), &tempnode, sizeof(UInt16) );
						forward1temp.left = avlarray[forward2].right;
						if( forward1temp.left != 0 )
							DmWrite( avlarray, forward1temp.left * sizeof(AVLNode) + 3*sizeof(UInt16), &forward1, sizeof(UInt16) );
					}

					forward2temp.prevnode = avlarray[tempnode].prevnode;
					if( forward2temp.prevnode != 0 )
						if( avlarray[forward2temp.prevnode].right == tempnode )
							DmWrite( avlarray, forward2temp.prevnode * sizeof(AVLNode) + 2*sizeof(UInt16), &forward2, sizeof(UInt16) );
						else
							DmWrite( avlarray, forward2temp.prevnode * sizeof(AVLNode) + sizeof(UInt16), &forward2, sizeof(UInt16) );
					forward1temp.prevnode = forward2;
					tempnodetemp.prevnode = forward2;
					
					DmWrite( avlarray, tempnode * sizeof(AVLNode), &tempnodetemp, sizeof(AVLNode) );
					DmWrite( avlarray, forward1 * sizeof(AVLNode), &forward1temp, sizeof(AVLNode) );
					DmWrite( avlarray, forward2 * sizeof(AVLNode), &forward2temp, sizeof(AVLNode) );
					
					temp1 = 0;
					DmWrite( avlarray, forward2 * sizeof(AVLNode), &temp1, sizeof(UInt16) );
					temp1 = getHeight( avlarray[forward1].right, avlarray) - getHeight( avlarray[forward1].left, avlarray ) ;
					DmWrite( avlarray, forward1 * sizeof(AVLNode), &temp1, sizeof(UInt16) );
					temp1 = getHeight( avlarray[tempnode].right, avlarray) - getHeight( avlarray[tempnode].left, avlarray ) ;
					DmWrite( avlarray, tempnode * sizeof(AVLNode), &temp1, sizeof(UInt16) );

					if( tempnode == head )
						head = forward2;
					break;
			}
		}
	}
	//finally, update the zero'th index with the new head and first free node indexes
	DmWrite( avlarray, sizeof(Int16), &head, sizeof( head ) );
	DmWrite( avlarray, 2*sizeof(Int16), &firstfreeindex, sizeof( firstfreeindex) );
	if( !isblockmode )		//only close the record if we are not in block mode, otherwise, leave it to the calling function to do that
	{
		MemHandleUnlock( myrecord );
		DmReleaseRecord( hashDb, 0, true ); 
	}
}

 
 
/* FUNCTION: changeDocPtrAtHashIndex                                                                       *
 * PURPOSE: Given a reference to the pre-opened hash database, an index of a hash record in the hash array *
 * and a document index value, this function will assign the document pointer of the hash record at the    *
 * specified index to equal the given document index.                                                      *
 * DATE: 7/23/02  UNTESTED!                                                                                *
 * CONSTANTS: 0 - the record containing the hash information                                               *
 * MODULES: PalmHashDb                                                                                     */
void changeDocPtrAtHashIndex( DmOpenRef hashDBref, UInt16 index, UInt16 newdocptr )
{
	MemHandle myRecord;
	AVLNode *avlarray;
	HashRecordType modifiedrecord;
		
	myRecord = DmGetRecord( hashDBref, 0 );				//get the record containing the hash array
	avlarray = MemHandleLock( myRecord );                        

	modifiedrecord.docindex = newdocptr;				//create a hash index to replace the existing one with the new document index pointer
	modifiedrecord.hashval = avlarray[index].data.hashval;
	
	//update the array with the new index
	DmWrite( avlarray, index * sizeof(AVLNode) + 4*sizeof(HashRecordType) , &modifiedrecord, sizeof(modifiedrecord) );	
	
	MemHandleUnlock( myRecord );
	DmReleaseRecord( hashDBref, 0, true );
}


/* FUNCTION: decrementDocPtrsGreaterThan
 * PURPOSE: Given a document index number, this function will go through the hash value array and *
 * decrement each document index that it finds whose value is higher than the given index. This   *
 * function is used to update the hash database pointers to the insertion order database, after a *
 * memo has been deleted. The function returns the hash index whose document pointer matches the  *
 * index passed to the function.                                                                  *
 * DATE: 7/23/02   UNTESTED!                                                                      *
 * CONSTANTS: 0 - the record number of the hash value array                                       *
 * MODULES PalmHashDb                                                                             *
 * SUGGESTIONS: Try to only call DmWrite once to write the whole array.                           *
 *              Shouldn't x be a larger data type than short?                                     */
UInt16 decrementDocPtrsGreaterThan( DmOpenRef hashDb, UInt16 docnumber )
{
	return(1);
}


/* FUNCTION: deleteHashIndex                                                              *
 * PURPOSE: This function will remove a specific index from the hash record array         *
 *          and bump down all the ones after it so that the array will not get fragmented *
 * DATE: 5/28/02   UNTESTED!                                                              *
 * CONSTANTS: 0 - record index in the hash database which contains the hash array         *
 * MODULES: PalmHashDb                                                                    */
void deleteHashIndex( UInt16 index, DmOpenRef hashDb )
{
}


/* FUNCTION: findIndexWithHashVal                                                             *
 * PURPOSE: Given a hash value and a reference to the pre-opened hash database, this fucntion *
 * searches through the hash AVL tree and returns the pointer to the index in the document    *
 * database of the matching hash value (it also returns, through a pointer parameter, the     *
 * index in the hash record in which the match was found).                                    *
 * DATE: 7/23/02                                                                              *
 * CONSTANTS: 0 - the record in the hash database that holds all the hash records             *
 * MODULES: PalmHashDb                                                                        *
 * SUGGESTION: The assumption here is that any hash value given to this function actually     * 
 * exists in the hash database, there is no error handling, but there should be.              */
UInt16 findIndexWithHashVal( UInt16 value, DmOpenRef hashDb, UInt16 *index )
{
	MemHandle myrecord;
	AVLNode *avlarray;
	UInt16 currentnode;
	
	myrecord = DmQueryRecord( hashDb, 0 );
	avlarray = MemHandleLock( myrecord );
	
	currentnode = avlarray[0].left;				//start at the head
	
	do											//go through the tree until we hit a null pointer (normal exit of this loop is through the "break")
	{
		if( value < avlarray[currentnode].data.hashval )				//if we are too high, go left
			currentnode = avlarray[currentnode].left;
		else if( value > avlarray[currentnode].data.hashval )			//if we are too low, go right
			currentnode = avlarray[currentnode].right;
		else															//if we are just right, make sure to update the index parameter, and get
		{																//out of the loop
			if( index != NULL )
				*index = currentnode;
			break;
		}
	} while ( currentnode != 0 );
	
	MemHandleUnlock( myrecord );
	
	return( avlarray[currentnode].data.docindex );
}


/* FUNCTION: findRecordWithDocIndex                                                           *
 * PURPOSE: Given an index value into the document database and a reference to the pre-opened *
 * hash database, this function will find and return the index of the first hash entry in the *
 * database that points to the same index in the document database.                           *
 * DATE: 7/23/02                                                                              *
 * CONSTANTS: 0 - the index of the record in the hash database that has all the hash values   *
 * MODULES: PalmHashDb                                                                        *
 * SUGGESTION: The assumption here is that any index value given to this function actually    * 
 * exists in the hash database, there is no error handling, but there should be.              */
UInt16 findRecordWithDocIndex( UInt16 docindex, DmOpenRef hashDb )
{
	UInt16 x;
	MemHandle hashrecord;
	AVLNode *hashrecorddata;

	hashrecord = DmQueryRecord( hashDb, 0 );
	hashrecorddata = MemHandleLock( hashrecord );

	for( x = 1; x < hashrecorddata[0].right; x++ )				//go through each valid index linearly
		if( docindex == hashrecorddata[x].data.docindex )		//stop when you find the match
			break;
			
	MemHandleUnlock( hashrecord );
	return( x );
}


/* FUNCTION: getHashValue                                                                      *
 * PURPOSE: Given a reference to the pre-opened hash database, and the index of the hash entry * 
 * in the hash array, this function will access the array and return the hash value at the     *
 * index. IT WILL NOT DO ERROR CHECKING OR BOUNDS CHECKING ON THE ARRAY! This function should  *
 * only be used if a previous hash database function has provided the index number.            *
 * DATE: 7/19/02                                                                               *
 * CONSTANTS: none                                                                             *
 * MODULES: PalmHashDb                                                                         */
UInt16 getHashValue( DmOpenRef hashDb, UInt16 index )
{
	MemHandle myrecord;
	AVLNode *avlarray;
	UInt16 hashvalue;
	
	myrecord = DmQueryRecord( hashDb, 0 );			//open the hash array record for reading
	avlarray = MemHandleLock( myrecord );
	
	hashvalue = avlarray[index].data.hashval;		//get the desired hash value
	
	MemHandleUnlock( myrecord );
	
	return( hashvalue );
}


/* FUNCTION: getHeight                                                                     *
 * PURPOSE: Given an index into the hash database, this function will find the "height" of *
 * the avlnode at that index.                                                              *
 * DATE: 7/20/02                                                                           */
UInt16 getHeight( UInt16 currentnode, AVLNode *avlarray )
{
	UInt16 height = 0;

	do		//have "currentnode" go down the tree, using the balance factors to determine the next step
	{
		if( currentnode == 0 )		//if the node isn't even valid
			break;

		switch( avlarray[currentnode].balance )
		{
		case -1:	currentnode = avlarray[currentnode].left; break;
		case 0:		currentnode = avlarray[currentnode].left; break;
		case 1:		currentnode = avlarray[currentnode].right; break;
		}

		height++;
	} while ( currentnode != 0 );

	return( height );
}


/* FUNCTION: insertHashIndex                                                                             *
 * PURPOSE: In the AVL implementation of the hash database, this function is merely a wrapper function   *
 * provided for compatibility with existing code. The actual workhorse function is the doInsertHashIndex *
 * function, which is called with the appropriate parameters from here.                                  *
 * DATE: 7/20/02                                                                                         *
 * MODULES: PalmHashDb                                                                                   */
void insertHashIndex( HashRecordType hashdata, DmOpenRef hashDb )
{
	doInsertHashIndex( hashdata, hashDb, 0, 0 );
}


/* FUNCTION: openHashDb2                                                 *
 * PURPOSE: Opens the hash values database, if it exists. If it doens't, *
 * then the function creates one large enough to hold NUMMEMOS hash      *
 * indexes of an array and initializes the zero'th index to hold the     *
 * a pointer to the first unused index of the array.                     *
 * DATE: 7/19/02                                                         *
 * CONSTANTS: "SyncHashDB" - name of the hash database                   *
 *            'sTrt' - 32bit creator ID of the database                  *
 *            NUMMEMOS                                                   *
 *            0 - the record in the database that holds the hash array   *
 * MODULES: PalmHashDb                                                   */
DmOpenRef openHashDb2()
{
	LocalID databaseid;
	DmOpenRef hashDb;
	AVLNode zeroindex, *temp;
	MemHandle hashrecord;
	UInt16 dummy = 0;
	
	databaseid = DmFindDatabase(0, "SyncHashDB");			//find the hash database
	if( !databaseid )										//if it is not found, then create it
	{
		zeroindex.right = 1;				//the "right" member of the zero'th index indicates the first unused index in the hash array
		zeroindex.left = 0;					//the "left" member of the zero'th index indicates the index which is the head of the AVL tree
		DmCreateDatabase(0, "SyncHashDB", 'sTrt', 0, false );
		databaseid = DmFindDatabase(0, "SyncHashDB");
		hashDb = DmOpenDatabase(0, databaseid, dmModeReadWrite );
		hashrecord = DmNewRecord( hashDb, &dummy, NUMMEMOS * sizeof(AVLNode) );		//now create the actual record
		temp = MemHandleLock( hashrecord );
		DmWrite( temp, 0, &zeroindex, sizeof( AVLNode ) );							//copy the zero'th index to the actual record
		MemHandleUnlock( hashrecord );
		DmReleaseRecord( hashDb, 0, true );
	}
	else
		hashDb = DmOpenDatabase(0, databaseid, dmModeReadWrite );
	
	return( hashDb );
}
#endif